@@ -1,3 +1,4 @@
+# frozen_string_literal: false
 require 'test/unit'
 class TestAssertion < Test::Unit::TestCase
   def test_wrong_assertion
@@ -5,4 +6,12 @@ class TestAssertion < Test::Unit::TestCase
     assert_match(/assertion message must be String or Proc, but TrueClass was given/, error.message)
     assert_match(/\A#{Regexp.quote(__FILE__)}:#{line}:/, error.backtrace[0])
   end
+
+  def test_timeout_separately
+    assert_raise(Timeout::Error) do
+      assert_separately([], <<~"end;", timeout: 0.1)
+        sleep
+      end;
+    end
+  end
 end
